unit WSockUtils;

interface
uses
  SysUtils, Classes, Windows, WinSock;

type
  EKTcpClient = class(Exception);
  TKTcpClient = class(TComponent)
  private
    // O
    FConnected: Boolean;
    FPort: Integer;
    FHost: string;
    FErrCode: Integer;
    FErrMsg: string;
    // Ŏgϐ
    FSocket: TSocket;
    FOnError: TNotifyEvent;
    FBufSize: Integer;
    FReadBuffer: string;
    function WSCheck(v: Integer; msg: string): Integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open; virtual;  // T[o[Ɛڑ
    procedure Close; virtual; // T[o[Ɛؒf
    // nIȃ\bh
    procedure send(s: string);
    function recv: string;
    // obt@Ǘ郁\bh
    procedure SendLn(cmd: string; eol: string = #13#10);
    function RecvLn: string;
    function RecvLnToDot: string;
    //
    property Host: string read FHost write FHost;
    property Port: Integer read FPort write FPort;
    property ReadBuffer: string read FReadBuffer write FReadBuffer;
    property ErrorCode: Integer read FErrCode;
    property ErrorMsg: string read FErrMsg;
    //
    property OnError: TNotifyEvent read FOnError write FOnError;
  end;

// T[o[IPAhX擾
function GetIpAddressStr(server: string): string;
// T[o[zXg擾
function GetHostEnt(server: string): PHostEnt;

// WinSock ̃G[𓾂
function GetSockErrorMsg(err: Integer): string;

var
  WSAData: TWSAData;

implementation

uses unit_string;

function GetSockErrorMsg(err: Integer): string;
begin
  case err of
    WSABASEERR:          Result := '[0] No Error';
    WSAEINTR:            Result := '[10004] Interrupted system call';
    WSAEBADF:            Result := '[10009] Bad file number';
    WSAEACCES:           Result := '[10013] Permission denied';
    WSAEFAULT:           Result := '[10014] Bad address';
    WSAEINVAL:           Result := '[10022] Invalid argument';
    WSAEMFILE:           Result := '[10024] Too many open files';
    WSAEWOULDBLOCK:      Result := '[10035] Operation would block';
    WSAEINPROGRESS:      Result := '[10036] Operation now in progress';
    WSAEALREADY:         Result := '[10037] Operation already in progress';
    WSAENOTSOCK:         Result := '[10038] Socket operation on non-socket';
    WSAEDESTADDRREQ:     Result := '[10039] Destination address required';
    WSAEMSGSIZE:         Result := '[10040] Message too long';
    WSAEPROTOTYPE:       Result := '[10041] Protocol wrong type for socket';
    WSAENOPROTOOPT:      Result := '[10042] Bad protocol option';
    WSAEPROTONOSUPPORT:  Result := '[10043] Protocol not supported';
    WSAESOCKTNOSUPPORT:  Result := '[10044] Socket type not supported';
    WSAEOPNOTSUPP:       Result := '[10045] Operation not supported on socket';
    WSAEPFNOSUPPORT:     Result := '[10046] Protocol family not supported';
    WSAEAFNOSUPPORT:     Result := '[10047] Address family not supported by protocol family';
    WSAEADDRINUSE:       Result := '[10048] Address already in use';
    WSAEADDRNOTAVAIL:    Result := '[10049] Can''t assign requested address';
    WSAENETDOWN:         Result := '[10050] Network is down';
    WSAENETUNREACH:      Result := '[10051] Network is unreachable';
    WSAENETRESET:        Result := '[10052] Net dropped connection or reset';
    WSAECONNABORTED:     Result := '[10053] Software caused connection abort';
    WSAECONNRESET:       Result := '[10054] Connection reset by peer';
    WSAENOBUFS:          Result := '[10055] No buffer space available';
    WSAEISCONN:          Result := '[10056] Socket is already connected';
    WSAENOTCONN:         Result := '[10057] Socket is not connected';
    WSAESHUTDOWN:        Result := '[10058] Can''t send after socket shutdown';
    WSAETOOMANYREFS:     Result := '[10059] Too many referencescan''t splice';
    WSAETIMEDOUT:        Result := '[10060] ڑ^CAEgBڑ悩̉ԓ͂܂B(Connection timed out)';
    WSAECONNREFUSED:     Result := '[10061] ڑ₳܂BڑPC̃|[gJĂȂ\܂B(Connection refused)';
    WSAELOOP:            Result := '[10062] Too many levels of symbolic links';
    WSAENAMETOOLONG:     Result := '[10063] File name too long';
    WSAEHOSTDOWN:        Result := '[10064] Host is down';
    WSAEHOSTUNREACH:     Result := '[10065] No Route to Host';
    WSAENOTEMPTY:        Result := '[10066] Directory not empty';
    WSAEPROCLIM:         Result := '[10067] Too many processes';
    WSAEUSERS:           Result := '[10068] Too many users';
    WSAEDQUOT:           Result := '[10069] Disc Quota Exceeded';
    WSAESTALE:           Result := '[10070] Stale NFS file handle';
    WSAEREMOTE:          Result := '[10071] Too many levels of remote in path';
    WSASYSNOTREADY:      Result := '[10091] Network SubSystem is unavailable';
    WSAVERNOTSUPPORTED:  Result := '[10092] WINSOCK DLL Version out of range';
    WSANOTINITIALISED:   Result := '[10093] Successful WSASTARTUP not yet performed';
    WSAHOST_NOT_FOUND:   Result := '[11001] Host not found';
    WSATRY_AGAIN:        Result := '[11002] Non-Authoritative Host not found';
    WSANO_RECOVERY:      Result := '[11003] Non-Recoverable errors: FORMER RREFUSED NOTIMP';
    WSANO_DATA:          Result := '[11004] Valid name no data record of requested type';
    else                 Result := '[' + IntToStr(err) + '] G[';
  end;
end;


// T[o[IPAhX擾
function GetIpAddressStr(server: string): string;
var
  phe: PHostEnt;
begin
  phe := GetHostEnt(server);
  if phe = nil then
  begin
    Result := ''; // s
  end else
  begin
    // in_addr^Axxx.xxx.xxx.xxx ̕ɕϊ
    Result := inet_ntoa(PInAddr(phe.h_addr_list^)^);
  end;
end;

function GetHostEnt(server: string): PHostEnt;
var
  addr: u_long;
  phe: PHostEnt;
begin
  // T[o[IPAhX𓾂
  phe := gethostbyname(PChar(server));
  if phe = nil then
  begin
    // IPAhX̎w肪H
    addr := inet_addr(PChar(server));
    phe  := gethostbyaddr(@addr, 4, AF_INET);
  end;
  Result := phe;
end;


{ TKTcpClient }

procedure TKTcpClient.Close;
begin
  if FSocket <> INVALID_SOCKET then
  begin
    WSCheck(shutdown(FSocket, SD_BOTH), 'ؒfɎs܂B');
    WSCheck(closesocket(FSocket), '\Pbg܂B');
    FConnected := False;
  end;
end;


constructor TKTcpClient.Create(AOwner: TComponent);
begin
  inherited;
  FConnected := False;
  FHost := '';
  FPort := 80; // Kɏ
  FSocket := INVALID_SOCKET;
  FBufSize := 8 * 1024;
  FReadBuffer := '';
end;

destructor TKTcpClient.Destroy;
begin
  Close;
  inherited;
end;

procedure TKTcpClient.Open;
var
  phe: PHostEnt;
  sockadd: sockaddr_in;
begin
  if FConnected then Close;

  // zXg𓾂
  phe := GetHostEnt(Host);
  if phe = nil then raise EKTcpClient.Create(Host + '݂͑܂B');

  // \Pbg
  FSocket := socket(PF_INET, SOCK_STREAM, 0);
  if(FSocket = INVALID_SOCKET)then
    raise EKTcpClient.Create('\Pbg쐬ł܂łB');

  // ڑAhXݒ
  ZeroMemory(@sockadd, Sizeof(sockadd));
  sockadd.sin_family := AF_INET;
  sockadd.sin_port   := htons(FPort);
  sockadd.sin_addr   := PInAddr(phe.h_addr_list^)^;

  // ڑ
  if WinSock.connect(FSocket, sockadd, sizeof(sockadd)) = SOCKET_ERROR then
  begin
    // ڑs
    FErrCode := WSAGetLastError;
    FErrMsg  := GetSockErrorMsg(FErrCode);
    raise EKTcpClient.Create('zXgƐڑł܂B' + FErrMsg);
  end;
  FConnected := True;
end;

function TKTcpClient.recv: string;
var
  buf: string;
  cnt: Integer;
begin
  Result := '';

  // M
  while True do
  begin
    // Mobt@m
    SetLength(buf, FBufSize);
    ZeroMemory(@buf[1], Length(buf));

    // M
    cnt := WinSock.recv(FSocket, buf[1], Length(buf)-1, 0);

    // Mł`FbN
    if cnt = SOCKET_ERROR then
    begin
      cnt := WSAGetLastError;
      raise EKTcpClient.Create('MG[:'+IntToStr(cnt));
    end;
    if cnt = 0 then Break;

    // obt@𒲐
    SetLength(buf, cnt);

    // Mʂ𓾂
    Result := Result + buf;
    if FBufSize > cnt then
    begin
      Break; // ǂ݂!!
    end;
  end;
end;

function TKTcpClient.RecvLn: string;
var
  i: Integer;
  s: string;
begin
  Result := '';

  while True do
  begin
    // [#13,#10]܂ޕ񂪂΂؂o
    i := Pos(#13, FReadBuffer);
    if i = 0 then i := Pos(#10, FReadBuffer);
    if i > 0 then
    begin
      // ʂ؂
      Result := Copy(FReadBuffer, 1, i - 1);
      System.Delete(FReadBuffer, 1, i - 1);

      // s폜Ĕ
      s := Copy(FReadBuffer, 1, 2);
      if s = #13#10 then
      begin
        System.Delete(FReadBuffer, 1, 2); Break;
      end;
      s := Copy(FReadBuffer, 1, 1);
      if (s = #13) or (s = #10) then
      begin
        System.Delete(FReadBuffer, 1, 1); Break;
      end;
    end;
    // 񂪂ȂΒǉœǂ
    FReadBuffer := FReadBuffer + recv;
  end;
end;

function TKTcpClient.RecvLnToDot: string;
var
  line: string;
begin
  Result := '';
  while True do
  begin
    line := RecvLn;
    if line = '.' then Break;
    Result := Result + line + #13#10;
  end;
end;

procedure TKTcpClient.send(s: string);
var
  err: Integer;
begin
  if s = '' then Exit;

  // M
  if WinSock.send(FSocket, s[1], Length(s), 0) = SOCKET_ERROR then
  begin
    err := WSAGetLastError;
    raise EKTcpClient.Create('MG[:'+IntToStr(err));
  end;
end;

procedure TKTcpClient.SendLn(cmd, eol: string);
begin
  Self.send(cmd + eol);
end;

function TKTcpClient.WSCheck(v: Integer; msg: string): Integer;
begin
  if v = SOCKET_ERROR then
  begin
    FErrCode := WSAGetLastError;
    FErrMsg  := msg + GetSockErrorMsg(FErrCode);
    if Assigned(FOnError) then FOnError(Self);
  end;
  Result := v;
end;

initialization
  WSAStartup($0101, WSAData); // WinSock 

finalization
  WSACleanup; // WinSock I

end.
